
import numpy as np
from matplotlib import rcParams
rcParams['font.family'] = 'sans-serif'
rcParams['font.sans-serif'] = ['Arial']
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator, LogLocator
from matplotlib.transforms import ScaledTranslation
import pandas as pd
import seaborn as sns

from nnfwtbn.process import Process
from nnfwtbn.cut import Cut
from nnfwtbn.variable import Variable
import nnfwtbn.error as err

ATLAS = "Work in Progress"
INFO = "$\sqrt{s} = 13\,\mathrm{TeV}$, $36.1\,\mathrm{fb}^{-1}$\nSome selection"

FONT_SIZE = 16
LABEL_FONT_SIZE = 16
SUB_FONT_SIZE = 12

def atlasify(atlas=False, subtext=None, enlarge=1.3, axes=None):
    """
    Applies the atlas style to the plot. If the atlas argument evaluates to
    True, the ATLAS badge is added. If the atlas argument is a non-empty
    string, the string is appended after the badge.

    If the subtext argument is given, this argument is added on the line
    below. Multiple lines are printed separately.
    
    The enlarge factor defines how fare the y-axis should be extended
    to accommodate for the badge.

    If the axes argument is None, the currently active axes is used.
    """
    pt = 1 / 72  # 1 point in inches
    if axes is None:
        axes = plt.gca()

    ylim = axes.get_ylim()
    y_upper = ylim[0] + (ylim[1] - ylim[0]) * enlarge
    axes.set_ylim((ylim[0], y_upper))

    axes.legend(frameon=False, loc=1)

    axes.tick_params("both", which="both", direction="in")
    axes.tick_params("both", which="major", length=6)
    axes.tick_params("both", which="minor", length=3)
    axes.tick_params("x", which="both", top=True)
    axes.tick_params("y", which="both", right=True)
    axes.xaxis.set_minor_locator(AutoMinorLocator())
    axes.yaxis.set_minor_locator(AutoMinorLocator())

    transBase = axes.transAxes + ScaledTranslation(10 * pt, -(5 + FONT_SIZE) * pt,
                                                   axes.figure.dpi_scale_trans)
    transSubText = ScaledTranslation(0 * pt, 0, axes.figure.dpi_scale_trans)
    if atlas:
        t_atlas = axes.text(0, 1, "ATLAS",
                            transform=transBase,
                            fontdict={"size": FONT_SIZE, "style": "italic",
                                      "weight": "bold", })
        t_width = t_atlas.get_window_extent(axes.figure.canvas.get_renderer()).width
        if isinstance(atlas, str):
            transLabel = ScaledTranslation(t_width * pt + (FONT_SIZE / 3) * pt , 0,
                                           axes.figure.dpi_scale_trans)
            axes.text(0, 1, atlas,
                      transform=transBase + transLabel,
                      fontdict={"size": LABEL_FONT_SIZE, })
        transSubText += ScaledTranslation(0 * pt, -1.2 * FONT_SIZE * pt,
                                          axes.figure.dpi_scale_trans)

    if isinstance(subtext, str):
        for line in subtext.split("\n"):
            axes.text(0, 1, line,
                      transform=transBase + transSubText,
                      fontdict={"size": SUB_FONT_SIZE, })
            transSubText += ScaledTranslation(0 * pt, -1.2 * SUB_FONT_SIZE * pt,
                                              axes.figure.dpi_scale_trans)



class HistogramFactory:
    """
    Short-cut to create multiple histogram with the same set of processes or
    in the same region.
    """

    def __init__(self, *args, **kwds):
        """
        Accepts any number of positional and keyword arguments. The arguments
        are stored internally and use default value for hist(). See __call__().
        """
        self.args = args
        self.kwds = kwds

    def __call__(self, *args, **kwds):
        """
        Proxy for method to hist(). The positional argument passed to hist()
        are the positional argument given to the constructor concatenated with
        the positional argument given to this method. The keyword argument for
        hist() is the union of the keyword arguments passed to the constructor
        and this method. The argument passed to this method have precedence.

        The method returns the return value of hist.
        """
        return hist(*self.args, *args, **dict(self.kwds, **kwds))

def hist(dataframe, variable, bins, stacks, selection=None,
         range=None, blind=None, axes=None, figure=None,
         weight=None, y_log=False, y_min=None, vlines=[],
         denominator=0, numerator=-1, ratio_label=None, diff=False,
         ratio_range=None,
         **kwds):
    """
    Creates a histogram of stacked processes. The first argument is the
    dataframe to operate on. The 'variable' argument defines the x-axis. The
    variable argument can be a Variable object or a string naming a column in
    the dataframe.

    The 'bins' argument can be an integer specifying the number of bins or a
    list with all bin boundaries. If it is an integer, the argument range is
    mandatory. The range argument must be a tuple with the lowest and highest
    bin edge. The properties of a Variable object are used for the x- and
    y-axis labels.

    Stacks can contain process objects or lists of process objects. All items
    in a list are stacked. The type attributes of processes are considered
    during plotting.

    The optional blind argument controls which process should be blinded. The
    argument can be a single process, a list of processes or None. By default,
    no process is blinded.

    If the figure argument is omitted, this method creates a new axes and
    figure. If axes only is omitted, the method creates a new axes from the
    figure. If a ratio plot is drawn, axes must a list if main, ratio plot if
    not omitted.

    The method returns (figure, axes) which were used during plotting. This
    might be identical to the figure and axes arguments. If a ratio plot is
    drawn, the axes return value is a list of main, ratio plot.

    The weight is used to weight the entries. Entries have unit
    weight if omitted. The argument can be a string name of a column or a
    variable object.

    If the y_log argument is set to True, the y axis will be logarithmic. The
    axis is enlarged on a logarithmic scale to make room for the ATLAS labels.
    The optional y_min argument can be used to set the lower limit of the y
    axis. The default is 0 for linear scale, and 1 for logarithmic scale.

    The option vlines can be used to draw vertical lines onto the histogram,
    e.g., a cut line. The argument should be an array, one item for each line.
    If the item is a number a red line will be drawn at that x-position. If it
    is a dict, it will take the item 'x' to determine the position, all other
    keywords are passed to matplotlibs plot method.`

    Any other keyword argument is used to define the style. If the keyword
    argument is a list, it's length must equal the number of stacks. If the
    list item is a list, it is cycled for every stack member. If the list
    member is not a list, this value is used for all members of the stack.
    If the keyword is not a list,  the property is passed directly to
    matplotlib. The only exception is 'histtype'. The histtype argument can be
    'step', 'stepfilled' or 'points'. In the former two cases matplotlibs hist
    method is used, in the latter case 'errorbar' is used.

    The numerator and denominator arguments control the content of the ratio
    plot. The arguments can be either an integer or a list of processes. If
    an argument is an integer, it is used as an index on the stack list. If
    an argument is a list of processes, these processes will be evaluated and
    used for the ratio plot. If either is None, no ratio plot will be drawn.
    The default draws the ratio between the first and the last stack. If a
    ratio plot is drawn, the axes argument is be None or a list of main, ratio
    axes.

    The ratio_label option controls the label of the ratio plot.

    The ratio_range argument control the y-range of the ratio plot. If set to
    None, it will scale automatically to include all points. The default is
    is None.

    If diff is set to True, The difference between the 'numerator' and the
    'denominator' is down instead of their ratio.

    The module constants ATLAS and INFO are passed to atlasify. Overwrite them
    to change the badges.
    """
    # Wrap column string by variable
    if isinstance(variable, str):
        variable = Variable(variable, variable)

    if weight is None:
        weight = Variable("unity", lambda d: variable(d) * 0 + 1)
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    squared_weight = Variable("squared weight", lambda d: weight(d)**2)

    draw_ratio = (denominator is not None) and (numerator is not None)

    # Disable ratio plots if there is only one stack 
    if len(stacks) == 1 and \
            isinstance(denominator, int) and \
            isinstance(numerator, int):
        draw_ratio = False

    # Handle axes, figure
    if figure is None or axes is None:
        if figure is None:
            provider = plt
        else:
            provider = figure

        if draw_ratio:
            subplot_kwds = {"gridspec_kw": {"height_ratios": [3, 1]},
                            "figsize": (5.75, 5.75),
                            "sharex": True}
            figure, (axes, axes_ratio) = provider.subplots(2, 1,
                                                        **subplot_kwds)

        else:
            figure, axes = provider.subplots(figsize=(5.5, 4))
    all_axes = [axes]
    if draw_ratio:
        all_axes.append(axes_ratio)

    # Handle selection
    if selection is None:
        selection = Cut(lambda d: variable(d) * 0 == 0)
    elif not isinstance(selection, Cut):
        selection = Cut(selection)

    # Handle range/bins
    equidistant_bins = False
    if range is not None:
        # Build bins
        if not isinstance(bins, int):
            raise err.InvalidBins("When range is given, bins must be int.")
        if not isinstance(range, tuple)  or len(range) != 2:
            raise err.InvalidProcessSelection("Range argument must be a "
                                              "tuple of two numbers.")
        bins = np.linspace(range[0], range[1], bins + 1)
        equidistant_bins = True

    # Check structure of kwds
    new_kwds = {}
    for kwd, value in kwds.items():
        if isinstance(value, list):
            if len(value) != len(stacks):
                raise ValueError("Length of %s must equal number of stacks."
                                 % repr(kwd))

            # Wrap properties for the whole stack in a list
            new_kwds[kwd] = [(_ if isinstance(_, list) else [_]) for _ in value] 

        else:
            # Single value for all stacks and all processes
            new_kwds[kwd] = [[value]]*len(stacks)
    kwds = new_kwds

    numerator_hist = None
    denominator_hist = None

    # Handle stack
    for i_stack, stack in enumerate(stacks):
        if isinstance(stack, Process):
            # Wrap single process
            stack = [stack]

        bottom = np.zeros(len(bins) - 1)
        sosw = np.zeros(len(bins) - 1)  # Sum of squared weights
        for i_process, process in enumerate(stack):
            sel = selection & process.selection

            # Prepare style
            process_kwds = {}
            for kwd, props in kwds.items():
                # Cycle through property stacks
                stack_props = props[i_stack]
                process_kwds[kwd] = stack_props[i_process % len(stack_props)]

            if "histtype" in process_kwds and process_kwds["histtype"] == "points":
                del process_kwds['histtype']

                defaults = {
                    'markersize': 4,
                    'fmt': 'o'
                }

                defaults.update(process_kwds)
                process_kwds = defaults

                n, _ = np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=weight(sel(dataframe)))

                bin_centers = (bins[1:] + bins[:-1]) / 2
                bin_widths = bins[1:] - bins[:-1]

                axes.errorbar(bin_centers, bottom + n, np.sqrt(n), bin_widths / 2,
                    label=process.label,
                    **process_kwds)

                sosw += n

            else:
                n, _, _ = axes.hist(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    bottom=bottom,
                    label=process.label,
                    weights=weight(sel(dataframe)),
                    **process_kwds)

                sosw += np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=squared_weight(sel(dataframe)))[0]
            bottom += n 

        # (end of processes loop within a stack)

        if isinstance(numerator, int) and stacks[numerator] == stack:
            numerator_hist = bottom
            numerator_error = np.sqrt(sosw)
        if isinstance(denominator, int) and stacks[denominator] == stack:
            denominator_hist = bottom
            denominator_error = np.sqrt(sosw)

        # Draw uncertainty band
        if (sosw > 0).any():
            band_lower = bottom - np.sqrt(sosw) / 2
            axes.hist(bins[:-1],
                      bins=bins, range=range,
                      bottom=band_lower,
                      weights=np.sqrt(sosw),
                      fill=False,
                      hatch='/////',
                      linewidth=0,
                      edgecolor="#666666")

    # Handle ratio plot
    if draw_ratio:
        if numerator_hist is None:
            if isinstance(numerator, Process):
                # Wrap single process
                numerator = [numerator]
            numerator_hist = np.zeros(len(bins) - 1)
            numerator_error = np.zeros(len(bins) - 1)
            for process in numerator:
                sel = selection & process.selection
                numerator_hist += np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=weight(sel(dataframe)))[0]

                numerator_error += np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=squared_weight(sel(dataframe)))[0]
            numerator_error = np.sqrt(numerator_error)
        if denominator_hist is None:
            if isinstance(denominator, Process):
                # Wrap single process
                denominator = [denominator]
            denominator_hist = np.zeros(len(bins) - 1)
            denominator_error = np.zeros(len(bins) - 1)
            for process in denominator:
                sel = selection & process.selection
                denominator_hist += np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=weight(sel(dataframe)))[0]

                denominator_error += np.histogram(
                    variable(sel(dataframe)),
                    bins=bins, range=range,
                    weights=squared_weight(sel(dataframe)))[0]
            denominator_error = np.sqrt(denominator_error)


        idx = (denominator_hist != 0)
        if diff:
            idx = (idx == idx)

        bin_centers = ((bins[1:] + bins[:-1]) / 2)[idx]
        bin_widths = (bins[1:] - bins[:-1])[idx]

        if diff:
            ratio = numerator_hist[idx] - denominator_hist[idx]
            ratio_error = numerator_error[idx]
        else:
            ratio = numerator_hist[idx] / denominator_hist[idx]
            ratio_error = numerator_error[idx] / denominator_hist[idx]

        axes_ratio.errorbar(bin_centers,
                            ratio,
                            ratio_error,
                            bin_widths / 2,
                            color='k',
                            markersize=4, fmt='o')

        if diff:
            rel_error = denominator_error
            band_lower = 0 - rel_error
        else:
            denominator_hist[~idx] = 1
            numerator_error[idx] = 0
            rel_error = denominator_error / denominator_hist
            band_lower = 1 - rel_error

        axes_ratio.hist(bins[:-1],
                        bins=bins, range=range,
                        bottom=band_lower,
                        weights=rel_error * 2,
                        fill=False,
                        hatch='/////',
                        linewidth=0,
                        edgecolor="#666666")
        if ratio_range is not None:
            axes_ratio.set_ylim(ratio_range)

        axes_ratio.set_ylabel(ratio_label)

    # Draw vertical lines
    for vline in vlines:
        try: 
            x = vline['x']
            kwds = {k: v for k, v in vline.items() if k != 'x'}
        except TypeError:
            x = vline
            kwds = {'color': 'r'}
        for ax in all_axes:
            ax.plot([x, x], ax.get_ylim(), **kwds)

    atlasify(ATLAS, INFO, enlarge=1.0, axes=axes)

    # Configure x-axis
    axes.set_xlim((bins.min(), bins.max()))

    # Configure y-axis
    _, y_max = axes.get_ylim()

    if y_log:
        y_min = y_min if y_min is not None else 1
        y_max *= 10**1.6
        axes.set_yscale('log')
    else:
        y_min = y_min if y_min is not None else 0
        y_max *= 1.6

    axes.set_ylim((y_min, y_max))

    axes.legend(frameon=False, loc=1)

    x_label_axes = axes_ratio if draw_ratio else axes
    if variable.unit is not None:
        x_label_axes.set_xlabel("%s in %s" % (variable.name, variable.unit),
                        horizontalalignment='right', x=0.95)
    else:
        x_label_axes.set_xlabel(variable.name,
                                horizontalalignment='right', x=0.95)

    if equidistant_bins:
        if variable.unit is not None:
            axes.set_ylabel("Events / %g %s" % (bins[1] - bins[0], variable.unit),
                            horizontalalignment='right', y=0.95)
        else:
            axes.set_ylabel("Events / %g" % (bins[1] - bins[0]),
                            horizontalalignment='right', y=0.95)
    else:
        axes.set_ylabel("Events / Bin", horizontalalignment='right', y=0.95)

    for ax in all_axes:
        ax.tick_params("both", which="both", direction="in")
        ax.tick_params("both", which="major", length=6)
        ax.tick_params("both", which="minor", length=3)
        ax.tick_params("x", which="both", top=True)
        ax.tick_params("y", which="both", right=True)
        ax.xaxis.set_minor_locator(AutoMinorLocator())

    if not y_log:
        axes.yaxis.set_minor_locator(AutoMinorLocator())


    if draw_ratio:
        figure.subplots_adjust(hspace=0.025)
        return figure, (axes, axes_ratio)
    else:
        return figure, axes

def confusion_matrix(df, x_processes, y_processes, x_label, y_label,
                     weight=None, axes=None, figure=None, **kwds):
    """
    Creates a confusion matrix.
    """
    # Wrap column string by variable
    if weight is None:
        weight = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    # Handle axes, figure
    if figure is None:
        figure, axes = plt.subplots()  #figsize=(5,5))
    elif axes is None:
        axes = figure.subplots()

    y_processes.reverse()
    data = np.zeros((len(y_processes), len(x_processes)))
    for i_x, x_process in enumerate(x_processes):
        x_df = x_process(df)
        total_weight = weight(x_df).sum()
        for i_y, y_process in enumerate(y_processes):
            x_y_df = y_process(x_df)
            data[i_y][i_x] = weight(x_y_df).sum() / total_weight

    data = pd.DataFrame(data,
                        columns=[p.label for p in x_processes],
                        index=[p.label for p in y_processes])
    sns.heatmap(data, **dict(vmin=0, vmax=1, cmap="Greens", ax=axes,
                           cbar_kws={
                               'label': "$P($%s$|$%s$)$" % (y_label, x_label)
                               }
                           ), **kwds)
    axes.set_xlabel(x_label)
    axes.set_ylabel(y_label)
    axes.set_ylim(len(y_processes), 0)
    figure.subplots_adjust(top=0.85)

    if ATLAS is not None:
        figure.text(0.18, 0.89, "ATLAS",
                  fontdict={"size": 18, "style": "italic", "weight": "bold"})
        if isinstance(ATLAS, str):
            figure.text(0.35, 0.89, ATLAS,
                      fontdict={"size": 12, })

    return data


def roc(df, signal_process, background_process, discriminant, steps=100,
        selection=None,
        min=None, max=None, axes=None, weight=None):
    """
    Creates a ROC.
    """
    # Wrap column string by variable
    if discriminant is None:
        discriminant = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(discriminant, str):
        discriminant = Variable(discriminant, discriminant)

    if weight is None:
        weight = Variable("unity", lambda d: np.ones(len(d)))
    elif isinstance(weight, str):
        weight = Variable(weight, weight)

    # Handle selection
    if selection is None:
        selection = Cut(lambda d: np.ones(len(d), dtype='bool'))
    elif not isinstance(selection, Cut):
        selection = Cut(selection)

    # Handle axes
    if axes is None:
        fig, axes = plt.subplots()
    if min is None:
        min = discriminant(df).min()
    if max is None:
        max = discriminant(df).max()

    df = selection(df)

    signal_effs = []
    background_ieffs = []
    n_total_sig = weight(signal_process.selection(df)).sum()
    n_total_bkg = weight(background_process.selection(df)).sum()
    for cut_value in np.linspace(min, max, steps):
        residual_df = df[discriminant(df) >= cut_value]

        n_total = weight(residual_df).sum()
        if n_total == 0:
            continue

        signal_df = signal_process.selection(residual_df)
        background_df = background_process.selection(residual_df)

        n_signal = weight(signal_df).sum()
        n_background = weight(background_df).sum()

        signal_effs.append(n_signal / n_total_sig)
        background_ieffs.append(1 - n_background / n_total_bkg)

    data = pd.DataFrame({"Signal efficiency": signal_effs,
                         "1 - Background efficiency": background_ieffs})
    sns.lineplot(x="Signal efficiency", y="1 - Background efficiency", data=data,
                 ax=axes, ci=None, label=discriminant.name)
    axes.plot([0, 1], [1, 0], color='gray', linestyle=':')

    axes.set_xlim((0, 1))
    axes.set_ylim((0, 1.3))
    axes.legend(loc=1, frameon=False)

    axes.tick_params("both", which="both", direction="in")
    axes.tick_params("both", which="major", length=6)
    axes.tick_params("both", which="minor", length=3)
    axes.tick_params("x", which="both", top=True)
    axes.tick_params("y", which="both", right=True)
    axes.xaxis.set_minor_locator(AutoMinorLocator())
    axes.yaxis.set_minor_locator(AutoMinorLocator())

    if ATLAS is not None:
        axes.text(0.04, 0.89, "ATLAS", transform=axes.transAxes,
                  fontdict={"size": 18, "style": "italic", "weight": "bold"})
        if isinstance(ATLAS, str):
            axes.text(0.25, 0.89, ATLAS, transform=axes.transAxes,
                      fontdict={"size": 12, })

    return data



