
import unittest

import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from nnfwtbn.variable import Variable
from nnfwtbn.cut import Cut
from nnfwtbn.process import Process
from nnfwtbn.plot import roc, confusion_matrix, atlasify, hist, \
                         HistogramFactory


class SurvivalTestCase(unittest.TestCase):
    """
    Test that calling the plotting methods does not cause a crash. 
    
    (Testing plotting methods is notoriously difficult.)
    """

    def test_roc(self):
        """
        Check that calling roc() does not crash.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4})

        p_signal = Process("Signal", range=(0, 0))
        p_bkg = Process("Background", range=(1, 1))

        try:
            roc(df, p_signal, p_bkg, 'reco_signal')
        except Exception:
            self.fail("Calling roc() failed.")

    def test_roc_return(self):
        """
        Check that the return valueis not None
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4})

        p_signal = Process("Signal", range=(0, 0))
        p_bkg = Process("Background", range=(1, 1))

        return_value = roc(df, p_signal, p_bkg, 'reco_signal')
        self.assertIsNotNone(return_value)


    def test_roc_custom_selection(self):
        """
        Check that calling roc() with a custom selection does not crash.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2})

        p_signal = Process("Signal", range=(0, 0))
        p_bkg = Process("Background", range=(1, 1))

        c_sr = Cut(lambda d: d.is_sr > 0)

        try:
            roc(df, p_signal, p_bkg, 'reco_signal', selection=c_sr)
        except Exception:
            self.fail("Calling roc() failed.")

    def test_roc_custom_axis(self):
        """
        Check that calling roc() with an existing axis does not crash.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4})

        p_signal = Process("Signal", range=(0, 0))
        p_bkg = Process("Background", range=(1, 1))

        fig, ax = plt.subplots()

        try:
            roc(df, p_signal, p_bkg, 'reco_signal', axes=ax)
        except Exception:
            self.fail("Calling roc() failed.")


    def test_confusion_matrix(self):
        """
        Check that calling confusion_matrix() does not raise an exception.
        """
        df = pd.DataFrame({
                "sick": [0, 1, 1, 0, 0],
                "positive": [0, 0, 1, 1, 0],
                "weight": [1, 1, 1, 1.7, 1],
            })

        positive = Cut(lambda d: d.positive == 1, label="+")
        negative = Cut(lambda d: d.positive == 0, label="-")
        sick = Cut(lambda d: d.sick == 1, label="sick")
        healthy = Cut(lambda d: d.sick == 0, label="healthy")

        try:
            data = confusion_matrix(df, 
                             [positive, negative], [sick, healthy],
                             "Test result", "Truth",
                             weight=Variable("weight", "weight"),
                             annot=True)
        except Exception:
            self.fail("Calling confusion_matrix() failed.")

    def test_confusion_matrix_return(self):
        """
        Check that the return value is not None.
        """
        df = pd.DataFrame({
                "sick": [0, 1, 1, 0, 0],
                "positive": [0, 0, 1, 1, 0],
                "weight": [1, 1, 1, 1.7, 1],
            })

        positive = Cut(lambda d: d.positive == 1, label="+")
        negative = Cut(lambda d: d.positive == 0, label="-")
        sick = Cut(lambda d: d.sick == 1, label="sick")
        healthy = Cut(lambda d: d.sick == 0, label="healthy")

        data = confusion_matrix(df, 
                         [positive, negative], [sick, healthy],
                         "Test result", "Truth",
                         weight=Variable("weight", "weight"),
                         annot=True)
        self.assertIsNotNone(data)

    def test_atlasify(self):
        """
        Check that calling atlasify() does not raise an error.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4})

        p_signal = Process("Signal", range=(0, 0))
        p_bkg = Process("Background", range=(1, 1))

        roc(df, p_signal, p_bkg, 'reco_signal')
        try:
            atlasify()
        except Exception:
            self.fail("Calling atlasify() failed.")


    def test_hist(self):
        """
        Check that calling hist() does not raise an exception.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_ggh = Process("$gg$F", range=(0, 0))
        p_vbfh = Process("VBF", range=(1, 1))

        try:
            hist(df,
                 v_higgs,
                 selection=c_vbf,
                 bins=20,
                 stacks=[[p_ggh, p_vbfh]],
                 range=(50, 450),
                 weight="weight",
                 color=[sns.color_palette("Blues")[:2]
                       + sns.color_palette()[1:]])
        except Exception:
            self.fail("Calling hist() failed.")

    def test_hist_return(self):
        """
        Check that  the return value is not None.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_ggh = Process("$gg$F", range=(0, 0))
        p_vbfh = Process("VBF", range=(1, 1))

        fig, axes = hist(df,
                         v_higgs,
                         selection=c_vbf,
                         bins=20,
                         stacks=[[p_ggh, p_vbfh]],
                         range=(50, 450),
                         weight="weight",
                         color=[sns.color_palette("Blues")[:2]
                             + sns.color_palette()[1:]])
        self.assertIsNotNone(fig)
        self.assertIsNotNone(axes)

    def test_hist_return_wo_ratio(self):
        """
        Check that the return value is not None if there is no ratio plot.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_ggh = Process("$gg$F", range=(0, 0))
        p_vbfh = Process("VBF", range=(1, 1))

        fig, axes = hist(df,
                         v_higgs,
                         selection=c_vbf,
                         bins=20,
                         stacks=[[p_ggh, p_vbfh]],
                         range=(50, 450),
                         weight="weight",
                         numerator=None,
                         color=[sns.color_palette("Blues")[:2]
                             + sns.color_palette()[1:]])
        self.assertIsNotNone(fig)
        self.assertIsNotNone(axes)

    def test_hist_single_stack(self):
        """
        Check that the method works with a single stack. The ratio plot should
        be omitted.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_vbfh = Process("VBF", range=(1, 1))

        try:
            hist(df,
                 v_higgs,
                 selection=c_vbf,
                 bins=20,
                 stacks=[p_vbfh],
                 range=(50, 450),
                 weight="weight",
                 color=[sns.color_palette("Blues")[:2]
                     + sns.color_palette()[1:]])
        except Exception:
            self.fail("Calling hist() failed.")

    def test_hist_facory(self):
        """
        Check that calling a HistogramFactory does not raise an exception.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_ggh = Process("$gg$F", range=(0, 0))
        p_vbfh = Process("VBF", range=(1, 1))

        hist_factory = HistogramFactory(df,
                                        selection=c_vbf,
                                        stacks=[[p_ggh, p_vbfh]],
                                        weight="weight",
                                        color=[sns.color_palette("Blues")[:2]
                                              + sns.color_palette()[1:]])
                        

        try:
            hist_factory(v_higgs, bins=20, range=(50, 450))
        except Exception:
            self.fail("Calling a HistogramFactory failed.")


    def test_hist_wo_ratio(self):
        """
        Check that creating a plot without ratio plot does not crash.
        """
        df = pd.DataFrame({"fpid": [0]*4 + [1]*4,
                           "reco_signal": [0, 1]*4,
                           "is_sr": [0] * 2 + [1]*4 + [0]*2,
                           "weight": [1, 2, 3, 4]*2})

        v_higgs = Variable(r"$H$", "reco_signal")
        c_vbf = Cut(lambda d: d.is_sr == 1)
        p_ggh = Process("$gg$F", range=(0, 0))
        p_vbfh = Process("VBF", range=(1, 1))

        try:
            hist(df,
                 v_higgs,
                 selection=c_vbf,
                 bins=20,
                 stacks=[[p_ggh, p_vbfh]],
                 range=(50, 450),
                 weight="weight",
                 numerator=None,
                 color=[sns.color_palette("Blues")[:2]
                       + sns.color_palette()[1:]])
        except Exception:
            self.fail("Calling hist() w/o ratio plot failed.")
